# !/usr/bin/python
# coding=utf-8
from PySide2 import QtWidgets, QtCore
from uitk.widgets.menu import Menu
from uitk.widgets.mixins.attributes import AttributesMixin
from uitk.widgets.mixins.text import RichText, TextOverlay


class CheckBox(QtWidgets.QCheckBox, AttributesMixin, RichText, TextOverlay):
    """ """

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent)

        # Set the initial style for rich text depending on the current state.
        self.set_checkbox_rich_text_style(self.isChecked())
        # Set the style on future state changes.
        self.stateChanged.connect(
            lambda state: self.set_checkbox_rich_text_style(state)
        )

        # Override built-ins
        self.text = self.richText
        self.setText = self.setRichText
        self.sizeHint = self.richTextSizeHint
        self.menu = Menu(self, mode="option")

        self.set_attributes(**kwargs)

    def set_checkbox_rich_text_style(self, state):
        """ """
        if self.has_rich_text:
            self.set_rich_text_style(textColor="black" if state > 0 else "white")

    def checkState(self):
        """Get the state of a checkbox as an integer value.
        Simplifies working with tri-state checkboxes.
        """
        if self.isTristate():
            state = {
                QtCore.Qt.CheckState.Unchecked: 0,
                QtCore.Qt.CheckState.PartiallyChecked: 1,
                QtCore.Qt.CheckState.Checked: 2,
            }
            return state[super().checkState()]
        else:
            return 1 if self.isChecked() else 0

    def setCheckState(self, state):
        """Set the state of a checkbox as an integer value.
        Simplifies working with tri-state checkboxes.

        Parameters:
            state (int)(bool): 0 or False: unchecked, 1 or True: checked.
                    If tri-state: 0: unchecked, 1: paritally checked, 2: checked.
        """
        if self.isTristate():
            s = {
                0: QtCore.Qt.CheckState.Unchecked,
                1: QtCore.Qt.CheckState.PartiallyChecked,
                2: QtCore.Qt.CheckState.Checked,
            }
            return super().setCheckState(s[state])
        else:
            self.setChecked(state)

    def hitButton(self, pos: QtCore.QPoint) -> bool:
        """Overridden method from QAbstractButton, used internally by Qt to decide whether a mouse press event
        should change the checkbox's state.

        This implementation extends the clickable area to the entire widget's bounds (including empty space),
        not just the checkbox and its label.

        Parameters:
            pos (QPoint): The position of the mouse event.

        Returns:
            bool: True if the position of the mouse event is within the widget's bounds, otherwise False.
        """
        return QtCore.QRect(QtCore.QPoint(0, 0), self.size()).contains(pos)

    def mousePressEvent(self, event):
        """Overridden method from QWidget to handle mouse press events.

        This implementation checks if the event is a right click or a left click.
        For a right click, it shows the context menu if one is defined.
        For a left click, it toggles the state of the checkbox.

        Parameters:
            event (QMouseEvent): The mouse event.

        Note:
            Other mouse events are passed to the parent class.
        """
        if event.button() == QtCore.Qt.RightButton:
            if self.menu:
                self.menu.show()

        if self.isTristate():
            # The next_state dictionary defines the order in which states should be cycled.
            next_state = {
                QtCore.Qt.CheckState.Unchecked: QtCore.Qt.CheckState.PartiallyChecked,
                QtCore.Qt.CheckState.PartiallyChecked: QtCore.Qt.CheckState.Checked,
                QtCore.Qt.CheckState.Checked: QtCore.Qt.CheckState.Unchecked,
            }
            # Change the checkbox's state.
            self.setCheckState(next_state[self.checkState()])
        else:
            super().mousePressEvent(event)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys
    from PySide2.QtCore import QSize

    # Return the existing QApplication object, or create a new one if none exists.
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    w = CheckBox(
        parent=None,
        setObjectName="chk000",
        setText="A Check Box <b>w/Rich TextMixin</b>",
        resize=QSize(125, 45),
        setWhatsThis="",
        setChecked=False,
        setVisible=True,
    )
    # w.show()

    sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace,
        then right-click them and select 'Promote to...'.

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote",
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtGui, QtWidgets
from uitk.widgets.mixins.attributes import AttributesMixin
from uitk.widgets.mixins.convert import ConvertMixin


class Region(QtWidgets.QWidget, AttributesMixin, ConvertMixin):
    """A custom QWidget that represents a region with a specified shape and size.
    Emits an on_enter signal when the mouse cursor enters the region.
    """

    on_enter = QtCore.Signal()
    on_leave = QtCore.Signal()

    def __init__(
        self,
        parent,
        position=(0, 0),
        size=(45, 45),
        shape=QtGui.QRegion.Ellipse,
        visible_on_mouse_over=False,
        **kwargs,
    ):
        """Initialize the Region widget.

        Parameters:
                parent (QtWidgets.QWidget): The parent widget for the Region instance.
                position (QPoint or tuple, optional): A tuple of (x, y) coordinates specifying the position of the center
                        of the region. Default is (0, 0).
                size (QSize or tuple, optional): A tuple of (width, height) specifying the size of the region. Default is (45, 45).
                shape (QRegion.Shape, optional): The shape of the region (default is QtGui.QRegion.Ellipse).
                visible_on_mouse_over (bool): Top level children are hidden when the mouse is not over the region.
                **kwargs: Additional keyword arguments to pass to the MainWindow. ie. setVisible=False or setMouseTracking=True
        """
        super().__init__(parent)

        position = self.to_qobject(position, QtCore.QPoint)
        size = self.to_qobject(size, QtCore.QSize)

        self.resize(size)
        self.setGeometry(self.rect())

        # Create a QRegion with the specified shape
        rect = QtCore.QRect(0, 0, size.width(), size.height())
        self.region = QtGui.QRegion(rect, shape)

        self.visible_on_mouse_over = visible_on_mouse_over

        self.move(self.mapFromGlobal(position - self.rect().center()))

        self.cursor_inside = False  # Track whether the cursor is inside the region.

        self.setVisible(True)
        self.set_attributes(**kwargs)

    @property
    def visible_on_mouse_over(self):
        """Get or set the visibility of the top-level children of the Region widget when the mouse cursor is over it.

        When set to True, the top-level children widgets will be hidden initially and will be shown when the mouse cursor
        enters the region. The children widgets will be hidden again when the cursor leaves the region.

        When set to False, the top-level children widgets will remain visible regardless of the mouse cursor's position.
        """
        return self._visible_on_mouse_over

    @visible_on_mouse_over.setter
    def visible_on_mouse_over(self, value):
        """Set the visibility of the top-level children of the Region widget when the mouse cursor is over it.

        Parameters:
                value (bool): If True, the top-level children widgets will be hidden initially and will be shown when the
                        mouse cursor enters the region. The children widgets will be hidden again when the cursor leaves the
                        region. Else, top-level children widgets will remain visible regardless of the mouse cursor's position.
        """
        if value:
            self.hide_top_level_children()
            self.on_enter.connect(self.show_top_level_children)
            self.on_leave.connect(self.hide_top_level_children)
        else:
            try:
                self.on_enter.disconnect(self.show_top_level_children)
                self.on_leave.disconnect(self.hide_top_level_children)
            except RuntimeError:
                pass
        self._visible_on_mouse_over = value

    def hide_top_level_children(self):
        """Hide all top-level child widgets of the Region instance."""
        for child in self.children():
            if isinstance(child, QtWidgets.QWidget):
                child.hide()

    def show_top_level_children(self):
        """Show all top-level child widgets of the Region instance."""
        for child in self.children():
            if isinstance(child, QtWidgets.QWidget):
                child.show()

    def enterEvent(self, event):
        """Overrides the QWidget.enterEvent method. Emits the on_enter signal when
        the cursor enters the Region widget's area, ensuring that the signal is
        emitted only once.

        Parameters:
                event (QEvent): The event object passed to the method.
        """
        cursor_pos = self.mapFromGlobal(QtGui.QCursor.pos())
        if not self.cursor_inside and self.rect().contains(cursor_pos):
            self.cursor_inside = True
            self.on_enter.emit()
        super().enterEvent(event)

    def leaveEvent(self, event):
        """Overrides the QWidget.leaveEvent method. Emits the on_leave signal when
        the cursor leaves the Region widget's area, ensuring that the signal is
        emitted only once.

        Parameters:
                event (QEvent): The event object passed to the method.
        """
        cursor_pos = self.mapFromGlobal(QtGui.QCursor.pos())
        if self.cursor_inside and not self.rect().contains(cursor_pos):
            self.cursor_inside = False
            self.on_leave.emit()
        super().leaveEvent(event)

    def hideEvent(self, event):
        """Overrides the QWidget.hideEvent method. Emits the on_leave signal when
        the widget is hidden.

        Parameters:
                event (QEvent): The event object passed to the method.
        """
        if self.visible_on_mouse_over:
            self.hide_top_level_children()
        self.cursor_inside = False
        super().hideEvent(event)

    def childEvent(self, event):
        """Overrides the QWidget.childEvent method. Hides the child widget if the
        visible_on_mouse_over flag is set to True.

        Parameters:
                event (QChildEvent): The event object passed to the method.
        """
        if self.visible_on_mouse_over and event.type() == QtCore.QEvent.ChildAdded:
            self.hide_top_level_children()

        super().childEvent(event)


# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    def on_region_enter():
        print("Mouse entered the region")

    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    main_window = QtWidgets.QMainWindow()
    main_window.setWindowTitle("Region Example")
    main_window.resize(300, 300)

    region_widget = Region(
        main_window, position=(150, 150), size=(50, 50), setMouseTracking=True
    )
    region_widget.on_enter.connect(on_region_enter)

    main_window.show()

    sys.exit(app.exec_())

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace, 
        then right-click them and select 'Promote to...'. 

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote", 
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""

# deprecated ---------------------



# !/usr/bin/python
# coding=utf-8
import os
import importlib
import pkgutil
import inspect
from .switchboard import signals  # Make signals accessible at package root


__package__ = "uitk"
__version__ = '0.7.8'
__path__ = [os.path.abspath(os.path.dirname(__file__))]


# Define a dictionary to map class names to their respective modules
CLASS_TO_MODULE = {}

# Build the CLASS_TO_MODULE dictionary by iterating over all submodules of the package
for importer, modname, ispkg in pkgutil.walk_packages(__path__, __name__ + "."):
    module = importlib.import_module(modname)
    for name, obj in module.__dict__.items():
        if inspect.isclass(obj):
            CLASS_TO_MODULE[obj.__name__] = modname

# Define a dictionary to store imported module objects
IMPORTED_MODULES = {}


def __getattr__(name):
    # Check if the requested attribute is a class we need to import
    if name in CLASS_TO_MODULE:
        module_name = CLASS_TO_MODULE[name]
        if module_name not in IMPORTED_MODULES:
            # If the module hasn't been imported yet, import it and add it to the dictionary
            module = importlib.import_module(module_name)
            IMPORTED_MODULES[module_name] = module
        else:
            module = IMPORTED_MODULES[module_name]
        # Return the requested class object from the module
        return getattr(module, name)

    # If the requested attribute is not a class we handle, raise an AttributeError
    raise AttributeError(f"module {__package__} has no attribute '{name}'")


# --------------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------

# The CLASS_TO_MODULE dictionary is built by iterating over all submodules of the package and checking if each submodule defines any classes.
# The resulting dictionary maps class names to their respective module names.

# When __getattr__ is called with a class name, it checks if the class name is in the CLASS_TO_MODULE dictionary.
# If it is, __getattr__ imports the corresponding module if it hasn't already been imported, and returns the requested class object from the module using getattr.

# Therefore, the classes are imported on-demand as they are requested, rather than all at once when the package is first imported.
# This can help reduce startup time and memory usage, especially for large packages with many modules and classes.

# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtWidgets
from uitk.widgets.mixins.text import RichText
from uitk.widgets.mixins.attributes import AttributesMixin


class OptionBox(QtWidgets.QPushButton, AttributesMixin, RichText):
    """A subclass of QPushButton designed to represent an option box that can wrap
    around another widget and display a context menu when clicked.

    Attributes:
        text (method): Overrides the QPushButton's text method with richText.
        setText (method): Overrides the QPushButton's setText method with setRichText.
        sizeHint (method): Overrides the QPushButton's sizeHint method with richTextSizeHint.
        wrapped_widget (QWidget): The widget that the option box wraps around.
    """

    def __init__(self, parent=None, **kwargs):
        QtWidgets.QPushButton.__init__(self, parent)
        """Instantiates the OptionBox with specific text, configures its size based
        on the parent's size, and sets its additional attributes.

        Parameters:
            box_text (str): The text to be displayed on the option box.
            **kwargs: Arbitrary keyword arguments for setting additional attributes.
        """
        # override built-ins
        self.text = self.richText
        self.setText = self.setRichText
        self.sizeHint = self.richTextSizeHint

        self.setStyleSheet("OptionBox {border: none;}")

        self.setText("⧉")
        self.set_attributes(**kwargs)

        # Connect the clicked signal to the show_menu slot
        self.clicked.connect(self.show_menu)

    def remove_border_for_widget(self, wrapped_widget):
        """Removes the border from the wrapped widget and the instance of this class.
        This is achieved by applying a class-specific style sheet to both widgets.

        Parameters:
            wrapped_widget (QWidget): The widget that is being wrapped by the option box.
        """
        wrapped_widget.setStyleSheet(
            wrapped_widget.__class__.__name__ + " {border: none;}"
        )
        self.setStyleSheet(self.__class__.__name__ + " {border: none;}")

    def wrap(self, wrapped_widget):
        """Wraps the option box around another widget. This involves creating a container
        widget, setting up a new layout, and placing the wrapped widget and the option
        box within that layout.

        Parameters:
            wrapped_widget (QWidget): The widget to be wrapped by the option box.
        """
        g_parent = wrapped_widget.parent()
        container = QtWidgets.QWidget(g_parent)
        container.setProperty("class", "withBorder")

        if g_parent.layout() is not None:
            g_parent.layout().replaceWidget(wrapped_widget, container)
            g_parent.layout().update()
        else:
            initial_pos = wrapped_widget.pos()
            container.move(initial_pos)

        self.setObjectName("{}_optionBox".format(wrapped_widget.objectName()))

        self.setMaximumSize(
            wrapped_widget.size().height(), wrapped_widget.size().height()
        )
        wrapped_widget.setMinimumSize(
            wrapped_widget.size().width() - wrapped_widget.size().height(),
            wrapped_widget.size().height(),
        )

        layout = QtWidgets.QHBoxLayout(container)
        layout.setContentsMargins(1, 1, 1, 1)
        layout.setSpacing(0)
        layout.addWidget(wrapped_widget, 0)
        layout.addWidget(self, 1)

        self.wrapped_widget = wrapped_widget
        self.remove_border_for_widget(wrapped_widget)
        container.setVisible(True)

        # Install wrapped widget's event filter onto the option box
        self.installEventFilter(self.wrapped_widget)

    def show_menu(self):
        """Shows the option menu if it contains items."""
        if self.menu.contains_items:
            if not self.wrapped_widget.isVisible():
                # If the wrapped widget is not visible, show at cursor pos.
                orig_pos = self.menu.position
                self.menu.position = "cursorPos"
                self.menu.show()
                self.menu.position = orig_pos
            else:
                self.menu.show()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    # return the existing QApplication object, or create a new one if none exists.
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    window = QtWidgets.QWidget()
    parent = QtWidgets.QPushButton(window)
    parent.setText("Button w/option Box")
    parent.resize(140, 21)

    option_box = OptionBox(
        setText='<hl style="color:red;">⛾</hl>',
    )
    option_box.wrap(parent)
    from PySide2.QtGui import QFont

    option_box.setFont(QFont("Arial", 12))  # Set the font size for the options

    option_a = QtWidgets.QAction("Option A", option_box)
    option_b = QtWidgets.QAction("Option B", option_box)

    option_box.addAction(option_a)
    option_box.addAction(option_b)
    option_box.menu.add([option_a, option_b])

    window.show()
    sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace,
        then right-click them and select 'Promote to...'.

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote",
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""

# Deprecated: --------------------



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtWidgets
from uitk.widgets.menu import Menu
from uitk.widgets.mixins.attributes import AttributesMixin
from uitk.widgets.mixins.text import RichText, TextOverlay


class PushButton(QtWidgets.QPushButton, AttributesMixin, RichText, TextOverlay):
    """ """

    def __init__(self, parent=None, **kwargs):
        QtWidgets.QPushButton.__init__(self, parent)

        # override built-ins
        self.text = self.richText
        self.setText = self.setRichText
        self.sizeHint = self.richTextSizeHint
        self.menu = Menu(self, mode="option")

        self.set_attributes(**kwargs)


# ----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys
    from PySide2.QtCore import QSize

    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(
        sys.argv
    )  # return the existing QApplication object, or create a new one if none exists.

    w = PushButton(
        parent=None,
        setObjectName="button_test",
        setText='<hl style="color:black;">A QPushButton <hl style="color:violet;"><b>with Rich Text</b></hl>',
        resize=QSize(125, 45),
        setWhatsThis="",
        # setVisible=True,
    )

    w.show()
    sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace, 
        then right-click them and select 'Promote to...'. 

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote", 
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""



# !/usr/bin/python
# coding=utf-8
import inspect
from PySide2 import QtCore, QtGui, QtWidgets
import pythontk as ptk
from uitk.widgets.optionBox import OptionBox
from uitk.widgets.draggableHeader import DraggableHeader
from uitk.widgets.mixins.style_sheet import StyleSheet
from uitk.widgets.mixins.attributes import AttributesMixin


class Menu(QtWidgets.QWidget, AttributesMixin, StyleSheet):
    """A custom Qt Widget that serves as a menu with additional features.

    The Menu class inherits from QtWidgets.QWidget and mixes in AttributesMixin and StyleSheet.
    It provides a customizable menu with features such as draggable headers and apply buttons.
    The menu can be positioned relative to the cursor, a specific coordinate, a widget, or its parent.

    Attributes:
        on_item_added (QtCore.Signal): Signal emitted when an item is added to the menu.
        on_item_interacted (QtCore.Signal): Signal emitted when an item in the menu is interacted with.
    """

    on_item_added = QtCore.Signal(object)
    on_item_interacted = QtCore.Signal(object)

    def __init__(
        self,
        parent=None,
        mode="context",
        position="cursorPos",
        min_item_height=None,
        max_item_height=None,
        fixed_item_height=None,
        add_draggable_header=True,
        **kwargs,
    ):
        """Initializes a Menu instance.

        Parameters:
            parent (QtWidgets.QWidget, optional): The parent widget. Defaults to None.
            position (str, optional): The position of the menu. Can be "right", "cursorPos", a coordinate pair, or a widget. Defaults to "right".
            min_item_height (int, optional): The minimum height of items in the menu. Defaults to None.
            max_item_height (int, optional): The maximum height of items in the menu. Defaults to None.
            fixed_item_height (int, optional): The fixed height of items in the menu. Defaults to None.
            add_draggable_header (bool, optional): Whether to add a draggable header to the menu. Defaults to True.
            **kwargs: Additional keyword arguments to set attributes on the menu.
        """
        super().__init__(parent)

        self.mode = mode
        self.position = position
        self.min_item_height = min_item_height
        self.max_item_height = max_item_height
        self.fixed_item_height = fixed_item_height
        self.add_draggable_header = add_draggable_header
        self.kwargs = kwargs
        self.widget_data = {}
        self.prevent_hide = False
        self.option_box = None

        self.setProperty("class", "translucentBgWithBorder")
        self.setWindowFlags(QtCore.Qt.Tool | QtCore.Qt.FramelessWindowHint)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.setMinimumWidth(147)

        self.init_layout()

        self.installEventFilter(self)
        if self.parent():
            self.parent().installEventFilter(self)
        self.set_attributes(**kwargs)

    def setCentralWidget(self, widget):
        """Set a widget as the central widget, replacing the current one."""
        current_central_widget = getattr(self, "_central_widget", None)
        if current_central_widget:
            current_central_widget.deleteLater()  # delete the current central widget
        self._central_widget = widget  # set the new central widget
        self._central_widget.setProperty("class", "centralWidget")  # for stylesheet
        self.layout.addWidget(self._central_widget)  # add it to the layout

    def centralWidget(self):
        """Return the central widget."""
        return self._central_widget

    def init_layout(self):
        """ """
        # Create a new layout with no margins
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(1)
        self.setLayout(self.layout)

        # Create a central widget and set it to the layout
        self.setCentralWidget(QtWidgets.QWidget(self))

        # Create a QVBoxLayout inside the central widget
        self.centralWidgetLayout = QtWidgets.QVBoxLayout(self._central_widget)
        self.centralWidgetLayout.setContentsMargins(12, 12, 12, 12)
        self.centralWidgetLayout.setSpacing(1)

        # Create a form layout inside the QVBoxLayout
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(1)

        if self.add_draggable_header:
            # Create DraggableHeader instance and add it to the central widget layout
            self.draggable_header = DraggableHeader(self)
            self.centralWidgetLayout.addWidget(self.draggable_header)

        # Add grid layout to the central widget layout
        self.centralWidgetLayout.addLayout(self.gridLayout)

    def get_all_children(self):
        children = self.findChildren(QtWidgets.QWidget)
        return children

    @property
    def contains_items(self):
        """Check if the QMenu contains any items."""
        return bool(self.gridLayout.count())

    @ptk.cached_property
    def apply_button(self):
        """Property that returns an apply button."""
        if not self.parent():
            return None

        button = QtWidgets.QPushButton("Apply")
        button.setToolTip("Execute the command.")
        button.released.connect(lambda: self.parent().clicked.emit())
        button.setFixedHeight(26)
        button.hide()

        return button

    def title(self):
        """ """
        self.draggable_header.text()

    def setTitle(self, title="") -> None:
        """Set the menu's title to the given string.
        If no title is given, the function will attempt to use the menu parents text.

        Parameters:
            title (str): Text to apply to the menu's header.
        """
        self.draggable_header.setText(title)

    def get_items(self):
        """Get all items in the list.

        Returns:
            list: A list of all QWidget items in the list.
        """
        items = [
            self.gridLayout.itemAt(i).widget() for i in range(self.gridLayout.count())
        ]

        return items

    def get_item(self, identifier):
        """Return a QAction or QWidgetAction by index or text.

        Parameters:
            identifier (int or str): If an int, treats it as an index. If a str, treats it as the text of the item.

        Raises:
            ValueError: If the identifier is not an integer (index) or string (text).

        Returns:
            QAction or QWidgetAction: The item found by the identifier.
        """
        items = self.get_items()

        if isinstance(identifier, int):  # get by index
            if identifier < 0 or identifier >= len(items):
                raise ValueError("Index out of range.")
            item = items[identifier]
        elif isinstance(identifier, str):  # get by text
            for i in items:
                if i.text() == identifier:
                    item = i
                    break
            else:
                raise ValueError("No item found with the given text.")
        else:
            raise ValueError(
                f"Expected an integer (index) or string (text), got '{type(identifier)}'"
            )

        return item

    def get_item_data(self, widget):
        """Get data associated with a widget in the list or its sublists.

        This method returns the data associated with the widget in the list or any sublist. If the widget is not found, it returns None.

        Parameters:
            widget (QtWidgets.QWidget): The widget to get the data for.

        Returns:
            Any: The data associated with the widget, or None if the widget is not found.
        """
        try:
            return self.widget_data.get(widget)
        except KeyError:
            return None

    def set_item_data(self, widget, data):
        """Set data associated with a widget in the list or its sublists.

        This method sets the data associated with a widget in the list. If the widget is not found, it does nothing.

        Parameters:
            widget (QtWidgets.QWidget): The widget to set the data for.
            data: The data to associate with the widget.
        """
        if widget in self.get_items():
            self.widget_data[widget] = data

    def remove_widget(self, widget):
        """Remove a widget from the layout."""
        self.gridLayout.removeWidget(widget)
        if widget in self.widget_data:
            del self.widget_data[widget]

    def clear(self):
        """Clear all items in the list."""
        # We're going backwards to avoid index errors.
        for i in reversed(range(self.gridLayout.count())):
            widget = self.gridLayout.itemAt(i).widget()
            if widget:
                self.gridLayout.removeWidget(widget)
                widget.setParent(None)
                widget.deleteLater()

        # Reset the widget_data dictionary
        self.widget_data = {}

    def add(self, x, data=None, row=None, col=0, rowSpan=1, colSpan=None, **kwargs):
        """Add an item or multiple items to the list.

        The function accepts a string, an object, or a collection of items (a dictionary, list, tuple, set, or map).

        Parameters:
            x (str, object, dict, list, tuple, set, map): The item or items to add.
            data: Data to associate with the added item or items. Default is None.
            row (int): The row index at which to add the widget. Default is the last row.
            col (int): The column index at which to add the widget. Default is 0.
            rowSpan (int): The number of rows the widget should span. Default is 1.
            colSpan (int): The number of columns the widget should span. Default is the total number of columns.
            **kwargs: Additional arguments to set on the added item or items.

        Returns:
            widget: The added widget or list of added widgets.
        """
        if isinstance(x, dict):
            return [self.add(key, data=val, **kwargs) for key, val in x.items()]

        elif isinstance(x, (list, tuple, set)):
            return [self.add(item, **kwargs) for item in x]

        elif isinstance(x, map):
            return [self.add(item, **kwargs) for item in list(x)]

        # get the widget from the value passed to x
        if isinstance(x, str):
            try:
                widget = getattr(QtWidgets, x)(self)
            except (AttributeError, TypeError):
                widget = QtWidgets.QLabel()
                widget.setText(x)

        elif isinstance(x, QtWidgets.QWidget) or (
            inspect.isclass(x) and issubclass(x, QtWidgets.QWidget)
        ):
            widget = x(self) if callable(x) else x

        else:
            raise TypeError(
                f"Unsupported item type: expected str, QWidget, a subclass of QWidget, or a collection (list, tuple, set, map, dict), but got '{type(x)}'"
            )

        widget.item_text = lambda i=widget: self.get_item_text(i)
        widget.item_data = lambda i=widget: self.get_item_data(i)

        # If no position is specified, place the widget at the last row and first column
        if row is None:
            row = 0
            while self.gridLayout.itemAtPosition(row, col) is not None:
                row += 1

        # If no span is specified, make the widget span across all columns
        if colSpan is None:
            colSpan = self.gridLayout.columnCount()

        self.gridLayout.addWidget(widget, row, col, rowSpan, colSpan)
        self.on_item_added.emit(widget)

        self.set_item_data(widget, data)

        if self.min_item_height is not None:
            widget.setMinimumHeight(self.min_item_height)
        if self.max_item_height is not None:
            widget.setMaximumHeight(self.max_item_height)
        if self.fixed_item_height is not None:
            widget.setFixedHeight(self.fixed_item_height)

        self.set_attributes(widget, **kwargs)
        widget.installEventFilter(self)
        # Add the widget as a menu attribute
        setattr(self, widget.objectName(), widget)

        self.resize(self.sizeHint())
        self.layout.invalidate()

        return widget

    def get_padding(widget):
        """Get the padding values around a widget.

        This method calculates the padding values (distance from content to frame boundary) for a widget in all four directions.

        Parameters:
            widget (obj): A widget object to get the padding values for.

        Returns:
            tuple: A tuple containing padding values (horizontal padding, vertical padding).
        """
        frame_geo = widget.frameGeometry()
        geo = widget.geometry()

        left_padding = geo.left() - frame_geo.left()
        right_padding = frame_geo.right() - geo.right()
        top_padding = geo.top() - frame_geo.top()
        bottom_padding = frame_geo.bottom() - geo.bottom()

        return (left_padding + right_padding, top_padding + bottom_padding)

    def sizeHint(self):
        """Return the recommended size for the widget.

        This method calculates the total size of the widgets contained in the layout of the ExpandableList, including margins and spacing.

        Returns:
            QtCore.QSize: The recommended size for the widget.
        """
        total_height = 0
        total_width = 0

        for i in range(self.layout.count()):
            widget = self.layout.itemAt(i).widget()
            if widget:
                total_height += widget.sizeHint().height() + self.layout.spacing()
                total_width = max(total_width, widget.sizeHint().width())

        # Adjust for layout's top and bottom margins
        total_height += (
            self.layout.contentsMargins().top() + self.layout.contentsMargins().bottom()
        )
        # Adjust for layout's left and right margins for width
        total_width += (
            self.layout.contentsMargins().left() + self.layout.contentsMargins().right()
        )

        return QtCore.QSize(total_width, total_height)

    def create_option_box(self):
        """ """
        self.option_box = OptionBox()
        self.option_box.menu = self
        self.option_box.wrap(self.parent())

    def toggle_visibility(self):
        """ """
        if self.isVisible():
            self.hide()
        elif self.contains_items:
            self.show()

    def center_on_cursor_position(self):
        """ """
        pos = QtGui.QCursor.pos()  # global position
        center = QtCore.QPoint(
            pos.x() - (self.width() / 2), pos.y() - (self.height() / 4)
        )
        self.move(center)  # center on cursor position.

    def is_position_a_coordinate(self):
        """ """
        return isinstance(self.position, (tuple, list, set, QtCore.QPoint))

    def move_to_coordinate(self):
        """ """
        if not isinstance(self.position, QtCore.QPoint):
            self.position = QtCore.QPoint(self.position[0], self.position[1])
        self.move(self.position)

    def is_position_a_widget(self):
        """ """
        return not isinstance(self.position, (type(None), str))

    def move_to_widget_position(self):
        """ """
        pos = getattr(self.positionRelativeTo.rect(), self.position)
        self.move(self.positionRelativeTo.mapToGlobal(pos()))

    @staticmethod
    def position_widget_relative_to_parent(parent, widget, position):
        """ """
        # Get dimensions of the parent and the widget
        parent_width = parent.width()
        parent_height = parent.height()
        widget_width = widget.width()
        widget_height = widget.height()

        pos_dict = {  # Define position options
            "right": (
                parent_width,  # x coordinate
                parent_height // 2 - widget_height // 2,  # y coordinate
            ),
            "left": (
                -widget_width,  # x coordinate
                parent_height // 2 - widget_height // 2,  # y coordinate
            ),
            "top": (
                parent_width // 2 - widget_width // 2,  # x coordinate
                -widget_height,  # y coordinate
            ),
            "bottom": (
                parent_width // 2 - widget_width // 2,  # x coordinate
                parent_height,  # y coordinate
            ),
            "center": (
                parent_width // 2 - widget_width // 2,  # x coordinate
                parent_height // 2 - widget_height // 2,  # y coordinate
            ),
        }

        # If the position is not one of the options, raise an error
        if position not in pos_dict:
            raise ValueError(f"Invalid position: {position}")

        # Get the global position of the parent widget
        global_pos = parent.mapToGlobal(QtCore.QPoint(0, 0))
        # Add the local coordinates of the new position
        new_pos = global_pos + QtCore.QPoint(*pos_dict[position])
        widget.move(new_pos)

    def hide_on_leave(self) -> None:
        """Hides the menu if the cursor is not within the menu's boundaries when the timer times out.
        This method is connected to the menu_timer's timeout signal.
        """
        if not self.rect().contains(QtGui.QCursor.pos()):
            self.hide()

    def leaveEvent(self, event) -> None:
        """ """
        self.hide()

        super().leaveEvent(event)

    def hide(self, force=False) -> None:
        """Sets the widget as invisible.
        Prevents hide event under certain circumstances.

        Parameters:
            force (bool): override prevent_hide.
        """
        if force or not self.prevent_hide:
            for w in self.get_items():
                try:
                    if w.view().isVisible():  # comboBox menu open.
                        return
                except AttributeError:
                    pass

            super().hide()

    def setVisible(self, state) -> None:
        """Called every time the widget is shown or hidden on screen."""
        if self.prevent_hide:  # invisible
            return

        super().setVisible(state)

    def showEvent(self, event) -> None:
        """ """
        # set menu position
        if self.position == "cursorPos":
            self.center_on_cursor_position()
        elif self.is_position_a_coordinate():
            self.move_to_coordinate()
        elif self.is_position_a_widget():
            self.move_to_widget_position()
        elif self.parent():
            self.position_widget_relative_to_parent(self.parent(), self, self.position)

        super().showEvent(event)

    def eventFilter(self, widget, event):
        """ """
        if event.type() == QtCore.QEvent.MouseButtonPress:
            if widget is self.parent():
                if self.mode == "context" and event.button() == QtCore.Qt.RightButton:
                    self.toggle_visibility()
                elif self.mode == "popup" and event.button() == QtCore.Qt.LeftButton:
                    self.toggle_visibility()

        if event.type() == QtCore.QEvent.Show:
            if widget is self.parent():
                if self.contains_items:
                    if self.mode == "option":
                        # Add apply button to the central widget layout
                        if hasattr(self.parent(), "clicked"):
                            self.centralWidgetLayout.addWidget(self.apply_button)
                            self.apply_button.show()
                        if self.option_box is None:
                            self.create_option_box()

        elif event.type() == QtCore.QEvent.MouseButtonRelease:
            if widget in self.get_items():
                self.on_item_interacted.emit(widget)

        return super().eventFilter(widget, event)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    # Return the existing QApplication object, or create a new one if none exists.
    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    menu = Menu(position="cursorPos", setTitle="Drag Me")

    # Grid layout example
    # a = menu.add(["Label A", "Label B"])
    a = menu.add("Label A")
    b = menu.add("Label B")
    c = menu.add("QDoubleSpinBox", set_spinbox_by_value=1.0, row=0, col=1)
    d = menu.add("QDoubleSpinBox", set_spinbox_by_value=2.0, row=1, col=1)

    menu.on_item_interacted.connect(lambda x: print(x))

    menu.set_style(theme="dark")

    menu.show()
    print(menu.get_items())
    sys.exit(app.exec_())

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace,
        then right-click them and select 'Promote to...'.

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote",
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""



# !/usr/bin/python
# coding=utf-8
from typing import Union
import re
from PySide2 import QtCore, QtWidgets
import pythontk as ptk


class StyleSheet(QtCore.QObject):
    """StyleSheet class is responsible for generating, modifying, and applying CSS style sheets for various Qt widgets.
    The class also provides utility functions to adjust the appearance of widgets based on their properties or conditions.
    The StyleSheet class offers multiple theme presets (e.g., 'light' and 'dark') and allows the user to create custom
    theme presets by providing a color value dictionary.

    Methods:
    - get_color_values(cls, theme="light", **kwargs): Return the colorValues dict with any of the bracketed
                placeholders replaced by the value of any given kwargs of the same name.
    - remove_leading_whitespace(s): Remove the same amount of leading whitespace from each line in the input string
                as present in the first line.
    - get_style_sheet(cls, widget_type=None, theme="light", **kwargs): Get the styleSheet for the given widget type.
    - set_style(cls, widget, ratio=6, theme="light", hide_menu_button=False, append_to_existing=False, **kwargs): Set the
                styleSheet for the given widgets.
    - adjust_padding(widget_type): Remove padding when the text length / widget width ratio is below a given amount.
    - hide_menu_button(widget_type): Set the menu button as transparent.
    """

    themes = {
        "light": {
            "MAIN_FOREGROUND": "rgb(255,255,255)",  # Bright white for better contrast
            "MAIN_BACKGROUND": "rgb(70,70,70)",  # Slightly darker than widget background
            "MAIN_BACKGROUND_ALPHA": "rgba(70,70,70,185)",
            "WIDGET_BACKGROUND": "rgb(125,125,125)",  # As per your request
            "BUTTON_PRESSED": "rgb(120,120,120)",  # Slightly darker than widget background
            "BUTTON_HOVER": "rgb(82,133,166)",
            "TEXT_COLOR": "rgb(255,255,255)",  # Bright white for better contrast
            "TEXT_CHECKED": "rgb(255,255,255)",  # Bright white for better contrast
            "TEXT_DISABLED": "rgba(150,150,150,175)",
            "TEXT_HOVER": "rgb(255,255,255)",  # Bright white for better contrast
            "TEXT_BACKGROUND": "rgb(70,70,70)",  # Same as main background
            "BORDER_COLOR": "rgb(40,40,40)",  # Slightly darker than widget background
            "HIGHLIGHT_COLOR": "rgb(255,255,190)",
            "DISABLED_BACKGROUND": "rgb(85,85,85)",
            "PROGRESS_BAR_COLOR": "rgb(0,160,208)",
        },
        "dark": {
            "MAIN_FOREGROUND": "rgb(200,200,200)",  # Lighter color for better contrast
            "MAIN_BACKGROUND": "rgb(90,90,90)",  # Darker than widget background
            "MAIN_BACKGROUND_ALPHA": "rgba(90,90,90,185)",
            "WIDGET_BACKGROUND": "rgb(60,60,60)",  # As per your request
            "BUTTON_PRESSED": "rgb(50,50,50)",  # Slightly darker than widget background
            "BUTTON_HOVER": "rgb(82,133,166)",
            "TEXT_COLOR": "rgb(220,220,220)",  # Lighter color for better contrast
            "TEXT_CHECKED": "rgb(255,255,255)",  # Bright white for better contrast
            "TEXT_DISABLED": "rgba(150,150,150,175)",
            "TEXT_HOVER": "rgb(255,255,255)",  # Bright white for better contrast
            "TEXT_BACKGROUND": "rgb(30,30,30)",  # Same as main background
            "BORDER_COLOR": "rgb(20,20,20)",  # Slightly darker than widget background
            "HIGHLIGHT_COLOR": "rgb(255,255,190)",
            "DISABLED_BACKGROUND": "rgb(35,35,35)",  # Slightly darker than main background
            "PROGRESS_BAR_COLOR": "rgb(0,160,208)",
        },
    }

    style_sheets = {
        "QWidget": """
            QWidget {
                background-color: transparent;
                border: none;
            }
            QWidget::item:selected {
                background-color: {BUTTON_HOVER};
            }
            QWidget.centralWidget {
                background-color: {MAIN_BACKGROUND_ALPHA};
                border: 1px solid {BORDER_COLOR};
                color: {TEXT_COLOR};
            }
            QWidget.QLabel {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_COLOR};
                border-style: outset;
                border-radius: 1px;
                border: 1px solid {BORDER_COLOR};
                padding: 0px 1px 0px 1px; /* top, right, bottom, left */
                spacing: 1px;
            }
        """,
        "QStackedWidget": """
            QStackedWidget {
                background-color: {MAIN_BACKGROUND};
            }
            QStackedWidget QFrame {
                background-color: {MAIN_BACKGROUND};
            }
        """,
        "QMainWindow": """
            QMainWindow {
                background-color: {MAIN_BACKGROUND_ALPHA};
                border: 1px solid {BORDER_COLOR};
            }
        """,
        "QGroupBox": """
            QGroupBox {
                border: 2px transparent;
                border-radius: 1px;
                padding: 1px 1px 1px 1px; /* top, right, bottom, left */
                margin: 12px 0px 1px 0px; /* top, right, bottom, left */ /* leave space at the top for the title */
                background-color: {MAIN_BACKGROUND_ALPHA};
            }
            QGroupBox::title {
                top: -13px;
                left: 2px;
                subcontrol-position: top left; /* position at the top center */
                background-color: {MAIN_BACKGROUND_ALPHA};
                color: {TEXT_COLOR};
            }
        """,
        "QMenu": """
            QMenu {
                background-color: {MAIN_BACKGROUND_ALPHA};
                border: 1px solid {BORDER_COLOR};
                margin: 0px; /* spacing around the menu */
            }
            QMenu::item {
                padding: 0px 1px 0px 1px; /* top, right, bottom, left */
                border: 1px solid transparent; /* reserve space for selection border */
            }
            QMenu::item:selected {
                border-color: {BUTTON_HOVER};
                background-color: {MAIN_BACKGROUND};
            }
            QMenu::icon:checked { /* appearance of a 'checked' icon */
                background-color: gray;
                border: 1px inset gray;
                position: absolute;
                top: 1px;
                right: 1px;
                bottom: 1px;
                left: 1px;
            }
            QMenu::separator {
                height: 2px;
                background-color: {MAIN_BACKGROUND};
                margin: 0px 5px 0px 10px; /* top, right, bottom, left */
            }
            QMenu::indicator {
                width: 13px;
                height: 13px;
            }
        """,
        "QMenuBar": """
            QMenuBar {
                background-color: {MAIN_BACKGROUND};
                spacing: 1px; /* spacing between menu bar items */
            }
            QMenuBar::item {
                padding: 1px 4px;
                background-color: transparent;
                border-radius: 1px;
            }
            QMenuBar::item:selected { /* when selected using mouse or keyboard */
                background-color: {BUTTON_HOVER};
            }
            QMenuBar::item:pressed {
                background-color: gray;
            }
        """,
        "QLabel": """
            QLabel {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_COLOR};
                border: 1px solid {BORDER_COLOR};
                border-radius: 1px;
                margin: 0px 0px 0px 0px; /* top, right, bottom, left */
                padding: 0px 2px 0px 2px; /* top, right, bottom, left */
            }
            QLabel::hover {
                border: 1px solid {BORDER_COLOR};
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
            QLabel::enabled {
                color: {TEXT_COLOR};
            }
            QLabel::disabled {
                color: {TEXT_DISABLED};
            }
        """,
        "QAbstractButton": """
            QAbstractButton {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_COLOR};
                border-style: outset;
                border-radius: 1px;
                border: 1px solid {BORDER_COLOR};
                padding: 0px 1px 0px 1px; /* top, right, bottom, left */
                spacing: 1px;
            }
            QAbstractButton::hover {
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
            QAbstractButton::hover:checked {
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
            QAbstractButton::enabled {
                color: {TEXT_COLOR};
            }
            QAbstractButton::disabled {
                color: {TEXT_DISABLED};
            }
            QAbstractButton::checked {
                background-color: {BUTTON_HOVER};
                color: {TEXT_CHECKED};
            }
            QAbstractButton::checked:hover {
                background-color: {BUTTON_HOVER};
                color: {TEXT_CHECKED};
            }
            QAbstractButton::indicator {
                width: 0px;
                height: 0px;
                border: none;
            }
            QAbstractButton::indicator::unchecked {
                image: none;
            }
            QAbstractButton::indicator:unchecked:hover {
                image: none;
            }
            QAbstractButton::indicator:unchecked:pressed {
                image: none;
            }
            QAbstractButton::indicator::checked {
                image: none;
            }
            QAbstractButton::indicator:checked:hover {
                image: none;
            }
            QAbstractButton::indicator:checked:pressed {
                image: none;
            }
        """,
        "QPushButton": """
            QPushButton {
                /* styles inherited from QAbstractButton */
            }
            QPushButton#toggle_expand {
                border: none;
                padding: 0px 0px 0px 0px; /* top, right, bottom, left */
                background-color: {MAIN_BACKGROUND};
            }
            QPushButton:flat {
                border: none; /* no border for a flat push button */
            }
            QPushButton:default {
                border-color: navy; /* make the default button prominent */
            }
        """,
        "QToolButton": """
            QToolButton {
                /* styles inherited from QAbstractButton */
            }
            QToolButton[popupMode="1"] { /* only for MenuButtonPopup */
                padding-right: 2px; /* make way for the popup button */
            }
            QToolButton:open { /* when the button has its menu open */
                background-color: dark gray;
            }
            /* popupMode set to DelayedPopup or InstantPopup */
            QToolButton::menu-indicator {
                image: none;
                subcontrol-origin: padding;
                subcontrol-position: bottom right;
                padding: 0px 5px 5px 0px; /* top, right, bottom, left */
            }
            QToolButton::down-arrow, QToolButton::up-arrow, QToolButton::left-arrow, QToolButton::right-arrow {
                image: none;
                padding: 0px 15px 0px 0px; /* top, right, bottom, left */
            }
            QToolButton::down-arrow:hover, QToolButton::up-arrow:hover, QToolButton::left-arrow:hover, QToolButton::right-arrow:hover {
                padding: 0px 5px 0px 0px; /* top, right, bottom, left */
            }
            /* the subcontrols below are used only in the MenuButtonPopup mode */
            QToolButton::menu-button {
                border: 1px solid {TEXT_COLOR};
                margin: 4px 2px 4px 0px; /* top, right, bottom, left */
            }
            QToolButton::menu-button:pressed {
                background-color: transparent;
                border: 1px solid transparent;
            }
            QToolButton::menu-arrow {
                image: none;
            }
        """,
        "QCheckBox": """
            QCheckBox {
                /* style inherited from QAbstractButton */
            }
        """,
        "QRadioButton": """
            QRadioButton {
                /* style inherited from QAbstractButton */
            }
        """,
        "QAbstractSpinBox": """
            QAbstractSpinBox {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_COLOR};
                border: 1px solid {BORDER_COLOR};
            }
            QAbstractSpinBox::disabled {
                color: {TEXT_DISABLED};
            }
            QAbstractSpinBox::hover {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_HOVER};
                border: 1px solid {BORDER_COLOR};
            }
            QScrollBar:left-arrow, QScrollBar::right-arrow, QScrollBar::up-arrow, QScrollBar::down-arrow {
                border: 1px solid {BUTTON_PRESSED};
                width: 3px;
                height: 3px;
            }
            QAbstractSpinBox::up-arrow, QAbstractSpinBox::down-arrow {
                width: 3px;
                height: 3px;
                border: 1px solid {BUTTON_PRESSED};
            }
            QAbstractSpinBox::up-button, QAbstractSpinBox::down-button {
                border: 1px solid {BUTTON_PRESSED};
                background-color: {WIDGET_BACKGROUND};
                subcontrol-origin: border;
            }
        """,
        "QSpinBox": """
            QSpinBox {
                /* style inherited from QAbstractSpinBox */
            }
        """,
        "QDoubleSpinBox": """
            QDoubleSpinBox {
                /* style inherited from QAbstractSpinBox */
            }
        """,
        "QComboBox": """
            QComboBox {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_COLOR};
                border: 1px solid {BORDER_COLOR};
                padding-left: 5px;
                border-radius: 1px;
                min-width: 0em;
            }
            QComboBox::hover {
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
            QComboBox::open {
                selection-background-color: {WIDGET_BACKGROUND};
                selection-color: {TEXT_CHECKED};
            }
            QComboBox:on { /* shift the text when the popup opens */
                padding-top: 3px;
            }
            QComboBox::down-arrow {
                width: 0px;
                height: 0px;
                background-color: {WIDGET_BACKGROUND};
                border: none;
                image: url(path/to/down_arrow.png);
            }
            QComboBox::drop-down {
                border: none;
                background-color: {WIDGET_BACKGROUND};
                subcontrol-origin: padding;
                subcontrol-position: top right;
                width: 0px;
                height: 0px;
                border-left-width: 1px;
                border-left-color: {TEXT_CHECKED};
                border-left-style: solid; /* just a single line */
                border-top-right-radius: 1px; /* same radius as the QComboBox */
                border-bottom-right-radius: 1px;
            }
            QComboBox QAbstractItemView {
                selection-background-color: {BUTTON_HOVER};
            }
        """,
        "QFrame": """
            QFrame {
                border-radius: 1px;
            }
        """,
        "QAbstractScrollArea": """
            QAbstractScrollArea {
                /* Inherits styles from QWidget */
                background-color: {WIDGET_BACKGROUND};
                border: 1px solid {BORDER_COLOR};
                selection-background-color: {BUTTON_HOVER};
                selection-color: {TEXT_COLOR};
            }
        """,
        "QScrollArea": """
            QScrollArea {
                /* Inherits styles from QAbstractScrollArea */
            }
        """,
        "QGraphicsView": """
            QGraphicsView {
                /* Inherits styles from QAbstractScrollArea */
            }
        """,
        "QMdiArea": """
            QMdiArea {
                /* Inherits styles from QAbstractScrollArea */
            }
        """,
        "QPlainTextEdit": """
            QPlainTextEdit {
                /* Inherits styles from QAbstractScrollArea */
            }
        """,
        "QTextEdit": """
            QTextEdit {
                background-color: {WIDGET_BACKGROUND};
                color: white;
                border: 1px solid {BORDER_COLOR};
                selection-background-color: {BUTTON_HOVER};
                selection-color: {TEXT_COLOR};
                background-attachment: fixed; /* fixed, scroll */
            }
            QTextEdit#hud_text {
                border: none;
                background-color: transparent;
                color: white;
                selection-background-color: {TEXT_BACKGROUND};
                selection-color: white;
            }
        """,
        "QLineEdit": """
            QLineEdit {
                background-color: {WIDGET_BACKGROUND};
                border: 1px solid {BORDER_COLOR};
                selection-background-color: {BUTTON_HOVER};
                selection-color: {TEXT_COLOR};
                border-radius: 1px;
                padding: 0 8px;
            }
            QLineEdit::disabled {
                color: {MAIN_BACKGROUND};
            }
            QLineEdit::enabled {
                color: {TEXT_COLOR};
            }
            QLineEdit:read-only {
                background-color: {MAIN_BACKGROUND};
            }
        """,
        "QAbstractItemView": """
            QAbstractItemView {
                alternate-background-color: {MAIN_BACKGROUND};
                background-attachment: fixed; /* fixed, scroll */
            }
            QAbstractItemView::item:alternate {
                background-color: {MAIN_BACKGROUND};
            }
            QAbstractItemView::item:selected {
                border: 1px solid {BORDER_COLOR};
                background-color: {BUTTON_HOVER};
                color: {TEXT_CHECKED};
            }
            QAbstractItemView::item:selected:!active {
                background-color: {BUTTON_HOVER};
                color: {TEXT_COLOR};
            }
            QAbstractItemView::item:selected:active {
                background-color: {BUTTON_HOVER};
                color: {TEXT_COLOR};
            }
            QAbstractItemView::item:hover {
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
        """,
        "QListWidget": """
            QListWidget {
                /* Inherits styles from QAbstractItemView */
            }
        """,
        "QListView": """
            QListView {
                /* Inherits styles from QAbstractItemView */
            }
        """,
        "QHeaderView": """
            QHeaderView {
                border: 1px solid {BUTTON_PRESSED};
            }
            QHeaderView::section {
                background-color: {BUTTON_PRESSED};
                border: 1px solid {BUTTON_PRESSED};
                padding: 1px;
            }
            QHeaderView::section:selected, QHeaderView::section::checked {
                background-color: {MAIN_BACKGROUND};
            }
        """,
        "QTableView": """
            QTableView {
                /* Inherits styles from QAbstractItemView */
            }
        """,
        "QTreeView": """
            QTreeView {
                /* Inherits styles from QAbstractItemView */
            }
            QTreeView QHeaderView {
                /* Inherits styles from QHeaderView */
            }
            QTreeView::branch {
                background: palette(base);
                border-image: none;
                border-width: 0;
                border-style: solid;
                border-color: {MAIN_BACKGROUND};
            }
        """,
        "QAbstractSlider": """
            QAbstractSlider {
                border: 1px solid black;
                background-color: {WIDGET_BACKGROUND};
            }
            QAbstractSlider::handle {
                width: 10px;
                margin-top: -6px;
                margin-bottom: -6px;
                border: 1px solid black;
                border-radius: 1px;
                background-color: {MAIN_BACKGROUND};
            }
            QAbstractSlider::handle:hover {
                background-color: {BUTTON_HOVER};
            }
            QAbstractSlider::groove {
                border: 1px solid black;
                height: 8px;
                background-color: {WIDGET_BACKGROUND};
            }
            QAbstractSlider::sub-page {
                background-color: {WIDGET_BACKGROUND};
            }
            QAbstractSlider::add-page {
                background-color: {WIDGET_BACKGROUND};
            }
        """,
        "QSlider": """
            /* Inherits styles from QAbstractSlider */
        """,
        "QScrollBar": """
            QScrollBar:vertical {
                background: {MAIN_BACKGROUND};
                width: 15px;
                margin: 15px 0 15px 0;
                border: 1px solid {BORDER_COLOR};
            }
            QScrollBar::handle:vertical {
                background: {WIDGET_BACKGROUND};
                min-height: 20px;
            }
            QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {
                border: 1px solid {BORDER_COLOR};
                background: {BUTTON_PRESSED};
                height: 15px;
                subcontrol-position: top;
                subcontrol-origin: margin;
            }
            QScrollBar::add-line:vertical:hover, QScrollBar::sub-line:vertical:hover {
                background: {BUTTON_HOVER};
            }
            QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical {
                background: none;
            }
            QScrollBar:horizontal {
                background: {MAIN_BACKGROUND};
                height: 15px;
                margin: 0 15px 0 15px;
                border: 1px solid {BORDER_COLOR};
            }
            QScrollBar::handle:horizontal {
                background: {WIDGET_BACKGROUND};
                min-width: 20px;
            }
            QScrollBar::add-line:horizontal, QScrollBar::sub-line:horizontal {
                border: 1px solid {BORDER_COLOR};
                background: {BUTTON_PRESSED};
                width: 15px;
                subcontrol-position: left;
                subcontrol-origin: margin;
            }
            QScrollBar::add-line:horizontal:hover, QScrollBar::sub-line:horizontal:hover {
                background: {BUTTON_HOVER};
            }
            QScrollBar::add-page:horizontal, QScrollBar::sub-page:horizontal {
                background: none;
            }
        """,
        "QProgressBar": """
            QProgressBar {
                border: 1px solid {BORDER_COLOR};
                border-radius: 1px;
                text-align: center;
                padding: 1px;
                background-color: {MAIN_BACKGROUND};
            }
            QProgressBar::chunk {
                background-color: {BUTTON_HOVER};
                width: 10px;
                margin: 0.5px;
            }
        """,
        # ... custom classes ...
        "default": """
            .default {
                background-color: {WIDGET_BACKGROUND};
                color: {TEXT_COLOR};
                border-style: outset;
                border-radius: 1px;
                border: 1px solid {BORDER_COLOR};
                padding: 0px 1px 0px 1px; /* top, right, bottom, left */
                spacing: 1px;
            }
            .default::hover {
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
            .default::hover:checked {
                background-color: {BUTTON_HOVER};
                color: {TEXT_HOVER};
            }
            .default::enabled {
                color: {TEXT_COLOR};
            }
            .default::disabled {
                color: {TEXT_DISABLED};
            }
        """,
        "translucentBgNoBorder": """
            .translucentBgNoBorder {
                background-color: rgba(127,127,127,0.005);
                border: none;
                color: {TEXT_COLOR};
                selection-background-color: {TEXT_BACKGROUND};
                selection-color: {HIGHLIGHT_COLOR};
            }
        """,
        "translucentBgWithBorder": """
            .translucentBgWithBorder {
                background-color: {MAIN_BACKGROUND_ALPHA};
                border: 1px solid {BORDER_COLOR};
                color: {TEXT_COLOR};
                selection-background-color: {TEXT_BACKGROUND};
                selection-color: {HIGHLIGHT_COLOR};
            }
        """,
        "noBorder": """
            .noBorder {
                border: 0px none;
            }
        """,
        "withBorder": """
            .withBorder {
                border: 1px solid {BORDER_COLOR};
            }
        """,
        "noPadding": """
            .noPadding {
                padding: 0px 0px 0px 0px; /* top, right, bottom, left */
            }
        """,
        "noHover": """
            .noHover:hover {
                background-color: {WIDGET_BACKGROUND};
            }
        """,
        "textBold": """
            .textBold {
                font-weight: bold;
            }
        """,
    }

    def get_style_sheet(self, widget_type=None, style="light", **kwargs):
        """Get the styleSheet for the given widget type.
        By default it will return all stylesheets as one multi-line css string.

        Parameters:
            widget_type (str): The class name of the widget. ie. 'QLabel'
            style (str): The color value set to use. valid values are: 'light', 'dark'

        Returns:
            (str) css styleSheet
        """
        css = (
            "".join(self.style_sheets.values())
            if widget_type is None
            else self.style_sheets.get(widget_type, "")
        )

        is_valid_widget_type = bool(getattr(QtWidgets, str(widget_type), None))
        if not css and is_valid_widget_type:
            raise ValueError(
                f"# Error: {__file__} in get_style_sheet\n#\tKeyError: '{widget_type}'"
            )
            return ""

        for k, v in self.get_color_values(style=style, **kwargs).items():
            css = css.replace(f"{{{k.upper()}}}", v)

        return self.remove_leading_whitespace(css)

    @staticmethod
    def get_super_type(widget_type):
        """Get the name of the immediate superclass of a widget type"""
        widget_class = getattr(QtWidgets, widget_type, None)
        if widget_class is not None:
            super_class = widget_class.__base__
            return super_class.__name__
        return None

    def get_style_hierarchy(self, widget_type, theme="light", **kwargs):
        """Recursively find and apply styles from the most abstract class to the specific class"""
        super_type = self.get_super_type(widget_type)

        # Try to get the style for the current widget type
        try:
            style = self.get_style_sheet(widget_type, theme=theme, **kwargs)
        except KeyError:
            style = ""

        # If this widget type has a superclass and it's not QWidget (since we already have a style for QWidget),
        # get the style of its superclass
        if super_type and super_type != "QWidget":
            try:
                super_style = self.get_style_sheet(super_type, theme=theme, **kwargs)
            except KeyError:
                super_style = ""
        else:
            super_style = ""

        return super_style, style

    def set_style(
        self,
        widget: Union[QtWidgets.QWidget, None] = None,
        theme="light",
        style_class="",
        **kwargs,
    ):
        """Set the styleSheet for the given widgets.
        Set the theme for a specific widget by using the '#' syntax and the widget's objectName. ie. QWidget#mainWindow

        Parameters:
            theme (str): Color mode. ie. 'light' or 'dark'
            widget (obj/list): The widget to set the theme of.
            style_class: Assign a custom style class to the widget.
        """
        if widget is None:
            widget = self

        if not isinstance(widget, QtWidgets.QWidget):
            raise ValueError(
                f"Invalid datatype for widget: expected QWidget, got {type(widget)}."
            )

        if style_class:
            widget.setProperty("class", style_class)

        # If the widget is a QMainWindow, apply the combined stylesheet
        if widget is self:
            final_style = self.get_style_sheet(theme=theme, **kwargs)
        else:  # Otherwise, apply the abstract style first, then the specific widget style
            widget_type = ptk.get_derived_type(
                widget, module="QtWidgets", return_name=True
            )
            super_style, style = self.get_style_hierarchy(
                widget_type, theme=theme, **kwargs
            )

            # Check for custom class style
            custom_style_class = widget.property("class")
            if custom_style_class and custom_style_class in self.style_sheets:
                style = self.get_style_sheet(custom_style_class, theme=theme, **kwargs)

            final_style = super_style + "\n" + style

        widget.setStyleSheet(final_style)

    @classmethod
    def get_color_values(cls, theme="light", **kwargs):
        """Return the colorValues dict with any of the bracketed placeholders
        replaced by the value of any given kwargs of the same name.

        Parameters:
            theme (str)(dict): The color value set to use. valid values are: 'light', 'dark'
                        or pass your own theme as a dict.
            **kwargs () = Keyword arguments matching the string of any bracketed placeholders.
                        case insensitive.  ex. alpha=255
        Returns:
            (dict) The color values with placeholder values. ex. {'MAIN_BACKGROUND_ALPHA': 'rgba(100,100,100,75)', etc..
        """
        if isinstance(theme, dict):
            return {
                k: v.format(**{k.upper(): v for k, v in kwargs.items()})
                for k, v in theme.items()
            }
        else:
            return {
                k: v.format(**{k.upper(): v for k, v in kwargs.items()})
                for k, v in cls.themes[theme].items()
            }

    @staticmethod
    def remove_leading_whitespace(style_sheet):
        """Remove the same amount of leading whitespace from each line in the input string as
        present in the first line.

        Parameters:
            style_sheet (str): Input string containing lines with leading whitespace.

        Returns:
            (str): Output string with leading whitespace removed from each line.
        """
        style_sheet = style_sheet.strip(
            "\n"
        )  # Remove leading and trailing newline characters
        lines = style_sheet.splitlines()
        if lines:
            leading_whitespace = re.match(r"^(\s*)", lines[0]).group(1)
            style_sheet = "\n".join(
                [re.sub(f"^{leading_whitespace}", "", line) for line in lines]
            )
        return style_sheet

    @staticmethod
    def adjust_padding(widget_type):
        """Remove padding when the text length / widget width ratio is below a given amount.

        Example:
            try:
                length = (  # a 'NoneType' error will be thrown if the widget does not contain text.
                    len(widget.text())
                    if hasattr(widget, "text")
                    else len(str(widget.value()))
                )
            except (AttributeError, ZeroDivisionError) as error:
                print (__name__.ljust(26), 'setStyle (adjustPadding)', widget.objectName().ljust(26), widget.__class__.__name__.ljust(25), error)

            # ratio of widget size, text length (using integer division).
            if widget.size().width() // length > ratio:
                final_style = final_style + self.adjust_padding(widget_type)
        """
        return """
            {0} {{
                padding: 0px 0px 0px 0px;
            }}""".format(
            widget_type
        )

    @staticmethod
    def hide_menu_button(widget_type):
        """Set the menu button as transparent."""
        return """
            {0}::menu-button {{
                border: 1px solid transparent;
            }}

            {0}::menu-button::hover {{
                border: 1px solid transparent;
            }}""".format(
            widget_type
        )


# --------------------------------------------------------------------------------------------

if __name__ == "__main__":
    pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------

"""
# css commenting:
    /* multi-line */

# setting a property:
    app.setStyleSheet("QLineEdit#name[prop=true] {background-color:transparent;}") #does't effect the lineEdit with objectName 'name' if that buttons property 'styleSheet' is false (c++ lowercase).
    widget.setProperty('prop', True)

# gradients:
    background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #E0E0E0, stop: 1 #FFFFFF);

# Set multiple widgets:
    QComboBox:!editable, QComboBox::drop-down:editable { ... }

# Set the style for a specific widget:
    QWidget#mainWindow ('#' syntax, followed by the widget's objectName)

# Add multiple child widgets to a single CSS rule by separating them with a space:
    QListWidget QPushButton, QListWidget QLabel { ... }

# Use the wildcard * to match any child widget of a widget, like this:
    QListWidget * { ... }


#qt styleSheet reference
http://doc.qt.io/qt-5/stylesheet-reference.html#qtabbar-widget


List of Pseudo-States:
:active             This state is set when the widget resides in an active window.
:adjoins-item       This state is set when the ::branch of a QTreeView is adjacent to an item.
:alternate          This state is set for every alternate row whe painting the row of a QAbstractItemView when QAbstractItemView::alternatingRowColors() is set to true.
:bottom             The item is positioned at the bottom. For example, a QTabBar that has its tabs positioned at the bottom.
:checked            The item is checked. For example, the checked state of QAbstractButton.
:closable           The items can be closed. For example, the QDockWidget has the QDockWidget::DockWidgetClosable feature turned on.
:closed             The item is in the closed state. For example, an non-expanded item in a QTreeView
:default            The item is the default. For example, a default QPushButton or a default action in a QMenu.
:disabled           The item is disabled.
:editable           The QComboBox is editable.
:edit-focus         The item has edit focus (See QStyle::State_HasEditFocus). This state is available only for Qt Extended applications.
:enabled            The item is enabled.
:exclusive          The item is part of an exclusive item group. For example, a menu item in a exclusive QActionGroup.
:first              The item is the first (in a list). For example, the first tab in a QTabBar.
:flat               The item is flat. For example, a flat QPushButton.
:floatable          The items can be floated. For example, the QDockWidget has the QDockWidget::DockWidgetFloatable feature turned on.
:focus              The item has input focus.
:has-children       The item has children. For example, an item in a QTreeView that has child items.
:has-siblings       The item has siblings. For example, an item in a QTreeView that siblings.
:horizontal         The item has horizontal orientation
:hover              The mouse is hovering over the item.
:indeterminate      The item has indeterminate state. For example, a QCheckBox or QRadioButton is partially checked.
:last               The item is the last (in a list). For example, the last tab in a QTabBar.
:left               The item is positioned at the left. For example, a QTabBar that has its tabs positioned at the left.
:maximized          The item is maximized. For example, a maximized QMdiSubWindow.
:middle             The item is in the middle (in a list). For example, a tab that is not in the beginning or the end in a QTabBar.
:minimized          The item is minimized. For example, a minimized QMdiSubWindow.
:movable            The item can be moved around. For example, the QDockWidget has the QDockWidget::DockWidgetMovable feature turned on.
:no-frame           The item has no frame. For example, a frameless QSpinBox or QLineEdit.
:non-exclusive      The item is part of a non-exclusive item group. For example, a menu item in a non-exclusive QActionGroup.
:off                For items that can be toggled, this applies to items in the "off" state.
:on                 For items that can be toggled, this applies to widgets in the "on" state.
:only-one           The item is the only one (in a list). For example, a lone tab in a QTabBar.
:open               The item is in the open state. For example, an expanded item in a QTreeView, or a QComboBox or QPushButton with an open menu.
:next-selected      The next item (in a list) is selected. For example, the selected tab of a QTabBar is next to this item.
:pressed            The item is being pressed using the mouse.
:previous-selected  The previous item (in a list) is selected. For example, a tab in a QTabBar that is next to the selected tab.
:read-only          The item is marked read only or non-editable. For example, a read only QLineEdit or a non-editable QComboBox.
:right              The item is positioned at the right. For example, a QTabBar that has its tabs positioned at the right.
:selected           The item is selected. For example, the selected tab in a QTabBar or the selected item in a QMenu.
:top                The item is positioned at the top. For example, a QTabBar that has its tabs positioned at the top.
:unchecked          The item is unchecked.
:vertical           The item has vertical orientation.
:window             The widget is a window (i.e top level widget)


List of Sub-Controls:
::add-line          The button to add a line of a QScrollBar.
::add-page          The region between the handle (slider) and the add-line of a QScrollBar.
::branch            The branch indicator of a QTreeView.
::chunk             The progress chunk of a QProgressBar.
::close-button      The close button of a QDockWidget or tabs of QTabBar
::corner            The corner between two scrollbars in a QAbstractScrollArea
::down-arrow        The down arrow of a QComboBox, QHeaderView (sort indicator), QScrollBar or QSpinBox.
::down-button       The down button of a QScrollBar or a QSpinBox.
::drop-down         The drop-down button of a QComboBox.
::float-button      The float button of a QDockWidget
::groove            The groove of a QSlider.
::indicator         The indicator of a QAbstractItemView, a QCheckBox, a QRadioButton, a checkable QMenu item or a checkable QGroupBox.
::handle            The handle (slider) of a QScrollBar, a QSplitter, or a QSlider.
::icon              The icon of a QAbstractItemView or a QMenu.
::item              An item of a QAbstractItemView, a QMenuBar, a QMenu, or a QStatusBar.
::left-arrow        The left arrow of a QScrollBar.
::left-corner       The left corner of a QTabWidget. For example, this control can be used to control position the left corner widget in a QTabWidget.
::menu-arrow        The arrow of a QToolButton with a menu.
::menu-button       The menu button of a QToolButton.
::menu-indicator    The menu indicator of a QPushButton.
::right-arrow       The right arrow of a QMenu or a QScrollBar.
::pane              The pane (frame) of a QTabWidget.
::right-corner      The right corner of a QTabWidget. For example, this control can be used to control the position the right corner widget in a QTabWidget.
::scroller          The scroller of a QMenu or QTabBar.
::section           The section of a QHeaderView.
::separator         The separator of a QMenu or in a QMainWindow.
::sub-line          The button to subtract a line of a QScrollBar.
::sub-page          The region between the handle (slider) and the sub-line of a QScrollBar.
::tab               The tab of a QTabBar or QToolBox.
::tab-bar           The tab bar of a QTabWidget. This subcontrol exists only to control the position of the QTabBar inside the QTabWidget. To style the tabs using the ::tab subcontrol.
::tear              The tear indicator of a QTabBar.
::tearoff           The tear-off indicator of a QMenu.
::text              The text of a QAbstractItemView.
::title             The title of a QGroupBox or a QDockWidget.
::up-arrow          The up arrow of a QHeaderView (sort indicator), QScrollBar or a QSpinBox.
::up-button         The up button of a QSpinBox.

List of Colors (Qt namespace (ie. Qt::red)):
white
black
red
darkRed
green
darkGreen
blue
darkBlue
cyan
darkCyan
magenta
darkMagenta
yellow
darkYellow
gray
darkGray
lightGray
color0 (zero pixel value) (transparent, i.e. background)
color1 (non-zero pixel value) (opaque, i.e. foreground)





image urls:
url(menu_indicator.png);
url(vline.png) 0;
url(handle.png);
url(close.png)
url(close-hover.png)
url(rightarrow.png);
url(leftarrow.png);
url(downarrow.png);
url(down_arrow.png);
url(up_arrow.png);
url(tear_indicator.png);
url(scrollbutton.png) 2;
url(branch-closed.png);
url(branch-open.png);
url(branch-more.png) 0;
url(branch-end.png) 0;
url(branch-open.png);
url(images/splitter.png);
url(images/splitter_pressed.png);
url(:/images/up_arrow.png);
url(:/images/up_arrow_disabled.png);
url(:/images/down_arrow.png);
url(:/images/down_arrow_disabled.png);
url(:/images/spindown.png) 1;
url(:/images/spindown_hover.png) 1;
url(:/images/spindown_pressed.png) 1;
url(:/images/sizegrip.png);
url(:/images/frame.png) 4;
url(:/images/spinup.png) 1;
url(:/images/spinup_hover.png) 1;
url(:/images/spinup_pressed.png) 1;
url(:/images/checkbox_unchecked.png);
url(:/images/checkbox_unchecked_hover.png);
url(:/images/checkbox_checked.png);
url(:/images/checkbox_checked_hover.png);
url(:/images/radiobutton_unchecked.png);
url(:/images/radiobutton_unchecked_hover.png);
url(:/images/radiobutton_checked.png);
url(:/images/radiobutton_checked_hover.png);
url(:/images/checkbox_indeterminate_hover.png);
url(:/images/checkbox_indeterminate_pressed.png);



"""

# min-width: 80px;

# QComboBox:editable {
#   background: {MAIN_BACKGROUND};
# }

# QComboBox:!editable, QComboBox::drop-down:editable {
#   background: {MAIN_BACKGROUND};
# }

# /* QComboBox gets the "on" state when the popup is open */
# QComboBox:!editable:on, QComboBox::drop-down:editable:on {
#   background: {MAIN_BACKGROUND};
# }

# background-color: #ABABAB; /* sets background of the menu */
# border: 1px solid black;

# /* sets background of menu item. set this to something non-transparent
# if you want menu color and menu item color to be different */
# background-color: transparent;

# QMenu::item:selected { /* when user selects item using mouse or keyboard */
#     background-color: #654321;
# For a more advanced customization, use a style sheet as follows:

# QMenu {
#     background-color: white;
#     margin: 2px; /* some spacing around the menu */

# QMenu::item {
#     padding: 2px 25px 2px 20px;
#     border: 1px solid transparent; /* reserve space for selection border */

# QMenu::item:selected {
#     border-color: darkblue;
#     background: rgba(100, 100, 100, 150);
# }

# QMenu::icon:checked { /* appearance of a 'checked' icon */
#     background: gray;
#     border: 1px inset gray;
#     position: absolute;
#     top: 1px;
#     right: 1px;
#     bottom: 1px;
#     left: 1px;

# QPushButton {
#   background:rgba(127,127,127,2);
#   background-color: red;
#   color: white;
#   border-width: 2px;
#   border-radius: 10px;
#   border-color: beige;
#   border-style: outset;
#   font: bold 14px;
#   min-width: 10em;
#   padding: 5px;

# QPushButton:hover {
#   border: 1px solid black;
#   border-radius: 5px;
#   background-color:#66c0ff;

# QPushButton:pressed {
#   background-color: rgb(224, 0, 0);
#   border-style: inset;

# QPushButton:enabled {
#   color: red

# QComboBox {
#   image: url(:/none);}

# QComboBox::drop-down {
#   border-width: 0px;
#   color: transparent

# QComboBox::down-arrow {
#   border: 0px solid transparent;
#   border-width: 0px; left: 0px; top: 0px; width: 0px; height: 0px;
#   background-color: transparent;
#   color: transparent;
#   image: url(:/none);

# QTreeView {
#   alternate-background-color: rgba(35,35,35,255);
#   background: rgba(45,45,45,255);
# }

# QMenu {
#   background-color: white; /* background-color: #ABABAB; sets background of the menu */
#   margin: 2px; /* some spacing around the menu */
#   border: 1px solid black;
# }

# QMenu::item {
#   /* sets background of menu item. set this to something non-transparent
#   if you want menu color and menu item color to be different */
#   background-color: transparent;
#   padding: 2px 25px 2px 20px;
#   border: 1px solid transparent; /* reserve space for selection border */
# }

# QMenu::item:selected {
#   /* when user selects item using mouse or keyboard */
#   background-color: #654321;
#   border-color: darkblue;
#   background: rgba(100, 100, 100, 150);
# }

# QMenu::icon:checked { /* appearance of a 'checked' icon */
#   background: gray;
#   border: 1px inset gray;
#   position: absolute;
#   top: 1px;
#   right: 1px;
#   bottom: 1px;
#   left: 1px;
# }

# /* non-exclusive indicator = check box style indicator (see QActionGroup::setExclusive) */
# QMenu::indicator:non-exclusive:unchecked {
#     image: url(:/images/checkbox_unchecked.png);
# }

# QMenu::indicator:non-exclusive:unchecked:selected {
#     image: url(:/images/checkbox_unchecked_hover.png);
# }

# QMenu::indicator:non-exclusive:checked {
#     image: url(:/images/checkbox_checked.png);
# }

# QMenu::indicator:non-exclusive:checked:selected {
#     image: url(:/images/checkbox_checked_hover.png);
# }

# /* exclusive indicator = radio button style indicator (see QActionGroup::setExclusive) */
# QMenu::indicator:exclusive:unchecked {
#     image: url(:/images/radiobutton_unchecked.png);
# }

# QMenu::indicator:exclusive:unchecked:selected {
#     image: url(:/images/radiobutton_unchecked_hover.png);
# }

# QMenu::indicator:exclusive:checked {
#     image: url(:/images/radiobutton_checked.png);
# }

# QMenu::indicator:exclusive:checked:selected {
#     image: url(:/images/radiobutton_checked_hover.png);


# QToolButton:hover, QToolButton::menu-button:hover {
#     background: #787876;
# }

# QToolButton::checked{
#     background: #484846;
#     border: 1px solid #787876;
# }

# QToolButton:pressed, QToolButton::menu-button:pressed {
#     background: #787876;
# }

# QToolButton[popupMode="1"]{
# /* only for MenuButtonPopup */
#     padding-right: 30px;
#     background: red;
# }
# QToolButton[popupMode="2"]{
# /* only for OSC Server Status */
#     padding-right: 30px;
#     background: #484846;
# }
# QToolButton[popupMode="2"]:hover{
#     background: #787876;
# }
# QToolButton::down-arrow{
# }
# /* the subcontrols below are used only in the MenuButtonPopup mode */
# QToolButton::menu-button{
# }

# QToolButton::menu-button:hover{
#     background: #787876;
# }
# QToolButton::menu-button:pressed{
# }
# QToolButton::menu-indicator{
#     bottom: 5px;
#     right: 5px;
# }



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtWidgets
from uitk.widgets.menu import Menu
from uitk.widgets.mixins.attributes import AttributesMixin
from uitk.widgets.mixins.text import RichText, TextOverlay


class ComboBox(QtWidgets.QComboBox, AttributesMixin, RichText, TextOverlay):
    """ """

    returnPressed = QtCore.Signal()
    beforePopupShown = QtCore.Signal()
    beforePopupHidden = QtCore.Signal()

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent)
        """ """
        self.menu = Menu(self, mode="option")

        self.view().installEventFilter(self)
        self.set_attributes(**kwargs)

    @property
    def items(self):
        """Get a list of each items's text or it's data if it exists from the standard model/view.

        Returns:
            (list)
        """
        return [
            self.itemData(i) if self.itemData(i) else self.itemText(i)
            for i in range(self.count())
        ]

    def block_signals(fn):
        """A decorator that blocks signals before executing a function, and unblocks them after.

        Parameters:
            fn (obj): The function to be decorated.
        """

        def wrapper(self, *args, **kwargs):
            self.blockSignals(True)  # prevent triggering currentIndexChanged
            rtn = fn(self, *args, **kwargs)
            self.blockSignals(False)
            return rtn

        return wrapper

    @block_signals
    def add(self, items, header=None, clear=True, ascending=False):
        """Add items to the combobox's standard modelView without triggering any signals.

        Parameters:
            items (str/list)(dict): A string, list of strings, or dict with 'string':data pairs to fill the comboBox with.
            header (str): An optional value for the first index of the comboBox's list.
            clear (bool): Clear any previous items before adding new.
            ascending (bool): Insert in ascending order. New item(s) will be added to the top of the list.

        Returns:
            (list) comboBox's current item list minus any header.

        Example:
            comboBox.add(["Import file", "Import Options"], "Import")
            comboBox.add({'Import file':<obj>, "Import Options":<obj>}, "Import") #example of adding items with data.
        """
        if not isinstance(items, (str, list, set, tuple, dict)):
            raise TypeError(
                f"Incorrect datatype. Expected str, list, set, tuple, or dict, got {type(items)}"
            )

        # Get the current index before refreshing list. avoid negative values.
        index = self.currentIndex() if self.currentIndex() > 0 else 0

        if clear:
            self.clear()

        if isinstance(items, str):
            items = [items]
        if isinstance(items, (list, tuple, set)):
            items = {i: None for i in items}

        for item, data in items.items():
            if item is not None:
                if ascending:
                    self.insertItem(0, item, data)
                else:
                    self.addItem(item, data)
        if header:
            self.insertItem(-1, header)

        self.setCurrentIndex(index)

        return items

    @block_signals
    def currentData(self):
        """Get the data at the current index.

        Returns:
            () data
        """
        index = self.currentIndex()
        return self.itemData(index)

    @block_signals
    def setCurrentData(self, value):
        """Sets the data for the current index.

        Parameters:
            value () = The current item's data value.
        """
        index = self.currentIndex()
        self.setItemData(index, value)

    @block_signals
    def currentText(self):
        """Get the text at the current index.

        Returns:
            (str)
        """
        index = self.currentIndex()
        return self.richText(index)

    @block_signals
    def setCurrentText(self, text):
        """Sets the text for the current index.

        Parameters:
            item (str): The current item's text value.
        """
        index = self.currentIndex()
        self.setRichText(text, index)

    @block_signals
    def setItemText(self, index, text):
        """Set the text at the given index.
        Override for setItemText built-in method.

        Parameters:
            item (str): Item text.
            index (int): Item index
        """
        self.setRichText(text, index)

    @block_signals
    def setCurrentItem(self, i):
        """Sets the current item from the given item text or index without triggering any signals.

        Parameters:
            i (str)(int): item text or item index
        """
        try:  # set by item index:
            self.setCurrentIndex(self.items.index(i))
        except Exception:  # set by item text:
            try:
                self.setCurrentText(i)
            except Exception as e:
                if i:
                    print(f"{__file__}: setCurrentItem: {e}")

    def showPopup(self):
        """Show the popup menu."""
        width = self.sizeHint().width()
        self.view().setMinimumWidth(width)

        self.beforePopupShown.emit()
        super().showPopup()

    def hidePopup(self):
        """ """
        self.beforePopupHidden.emit()
        super().hidePopup()

    def clear(self):
        """ """
        super().clear()

    def keyPressEvent(self, event):
        """ """
        if event.key() == QtCore.Qt.Key_Return and not event.isAutoRepeat():
            self.returnPressed.emit()
            self.setEditable(False)

        super().keyPressEvent(event)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(
        sys.argv
    )  # return the existing QApplication object, or create a new one if none exists.

    cmb = ComboBox(popupStyle="qmenu")
    cmb.show()
    sys.exit(app.exec_())


# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

"""
Promoting a widget in designer to use a custom class:
>   In Qt Designer, select all the widgets you want to replace,
        then right-click them and select 'Promote to...'.

>   In the dialog:
        Base Class:     Class from which you inherit. ie. QWidget
        Promoted Class: Name of the class. ie. "MyWidget"
        Header File:    Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>   Then click "Add", "Promote",
        and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
"""

# deprecated ---------------------
